open Support
open Support.Error
open Syntax

(* NB: we dump all the separation information. *)

(* To do: tail call opt *)

module Context = Map.Make(struct type t = variable let compare = String.compare end)
type context = typ Context.t

let rec string_typ: typ -> string =
  fun t ->
    match t.v with
    | UnitTyp     -> "()"
    | DiamondTyp  -> "<>"
    | BoolTyp     -> "bool"
    | IntTyp      -> "int"
    | FloatTyp    -> "float"
    | ArrayTyp(s) -> ("array["^(string_typ s)^"]")
    | CharTyp     -> "char"
    | StringTyp   -> "string"
    | TvarTyp(v)  -> ("'"^v) 
    | ConTyp(p,id) -> ((String.concat "_" (List.map string_typ p))^"_"^id)
    | ArrowTyp(d,r) -> ((string_typ d)^"->"^(string_typ r))
    | PairTyp(t1,t2) -> ((string_typ t1)^"__"^(string_typ t2))

let zero_tycon = function (TypCon (i,c,_,ts)) -> TypCon (i,c,Specified 0,ts)
let zero_tydec = function (TypDec (i,vs,t,cs)) ->
                             TypDec (i,vs,t,List.map zero_tycon cs)


let cps (Program (errinfo, tydecs, valdecs, fundefs)) mainfun =
  let contab = compile_contab tydecs in
  let stacktype = {i=unknown; v=ConTyp ([], "c%stack")} in

  (* Type reconstruction *)
  let rec typeofval (context, vl) =
    let i pty = {i=unknown; v=pty} in
      match vl.v with
      | VarVal v    -> Context.find v context
      | IntVal _    -> i IntTyp
      | FloatVal _  -> i FloatTyp
      | CharVal _   -> i CharTyp
      | StringVal _ -> i StringTyp
      | BoolVal _   -> i BoolTyp
      | UnitVal     -> i UnitTyp
      | UnaryOpVal (o,_) ->
          (match o.v with
           | NotOp     -> i BoolTyp
           | UMinusOp  -> i IntTyp
           | UFminusOp -> i FloatTyp)
      | BinaryOpVal (o,_,_) ->
          (match o.v with
           | TimesOp  | DivOp  | PlusOp  | MinusOp          -> i IntTyp
           | FtimesOp | FdivOp | FplusOp | FminusOp         -> i FloatTyp
           | AndOp  | OrOp   | AndalsoOp | OrelseOp | ModOp
           | LessOp | LteqOp | GreaterOp | GteqOp | EqualOp -> i BoolTyp
           | AppendOp -> i StringTyp
           | ConsOp -> failwith "ConsOp in code, but appears to be out-of-use!")
      | PairVal (v1,v2) ->
          i (PairTyp (typeofval (context, v1), typeofval (context, v2)))
  in

  let fnty f =
    match get_val_for f valdecs with
    | ValDec (_,_,_,ty)
        -> ty
    | AnnValDec (_,_,_,_,rty,_)
    | FullValDec (_,_,_,_,rty,_,_)
        -> strip_rich_typ rty
  in

  let rec fntys ty =
    match ty.v with
    | ArrowTyp (d,r) -> let (args, res) = fntys r in (d::args,res)
    | _ -> ([],ty)
  in

  let rec make_arrowty args final =
    match args with
    |  []  -> final
    | h::t -> {i=unknown; v=ArrowTyp (h,make_arrowty t final)}
  in

  let resultty ty =
    let (_, r) = fntys ty in r
  in

  (* It's a bit wasteful to recompute types at every let, but never mind. *)
  let rec typeof (context, expr) : typ =
    match expr.v with
    | ValueExp vl -> typeofval (context, vl)
    | ConstrExp (c,args,d) ->
        (* I don't think the type variables are used *)
        (ConTab.find c contab).own_typ
    | AppExp (f,_,_) -> resultty (fnty f)
    | LetExp (v,e1,e2,_) ->
        let ty1 : typ = {i=unknown; v=IntTyp} (*typeof (context,e1)*) in
        typeof (Context.add v ty1 context, e2)
    | SeqExp (_,e2) ->
        typeof (context, e2)
    | IfExp (_,e1,_) ->
        typeof (context, e1)
    | MatchPExp (v,v1,v2,e) ->
        let t1,t2 = match (Context.find v context).v with
                      PairTyp(t1,t2) -> t1,t2
                    | _ -> failwith "pair match given non-pair variable"
         in typeof (Context.add v2 t2 (Context.add v1 t1 context), e)
    | MatchExp (v,r::_) ->
        (match r with
         | Matchrule (_, c, bound, _, e, _) ->
             let boundtys = (ConTab.find c contab).arg_types in
             let context' = List.fold_left (fun c (v,t) -> Context.add v t c)
                              context (List.combine bound boundtys) in
             typeof (context', e))
  in

  (* We need to know the final type for all the function types. *)
  let finaltype =
    try resultty (fnty mainfun)
    with Not_found ->
      failwith ("Unable to find main function (" ^ mainfun ^ ") valdec")
  in

  let varval (var:variable) : value = {i=unknown; v=VarVal var} in

  let framesize f =
    match Syntax.get_val_for f valdecs with
      ValDec (_,_,s,_) | AnnValDec (_,_,s,_,_,_) | FullValDec (_,_,s,_,_,_,_) -> s
  in

  (* The CPS transform itself *)

  let rec exp (context, expr, cont, contvar, tail, origf) =
    match expr.v with
    |  ValueExp _ -> ({i=expr.i; v=LetExp (contvar, expr, cont, [])}, [])
    | ConstrExp _ -> ({i=expr.i; v=LetExp (contvar, expr, cont, [])}, [])
    | AppExp (f, args, appid) ->
        if tail then ({i=expr.i; v=AppExp (f, args@[varval "c%stack"], appid)}, []) else
        let fvcont = List.filter (fun v -> v <> contvar && v <> "c%stack")
                       (Syntax.fv_expr cont) in
        let fty = resultty (fnty f) in
        let contname = "c%" ^ (string_of_int appid) in
        let closname = "C%" ^ (string_of_int appid) in
        let clossize = if (!Argument.the_options).Argument.cpsclossize
                              (* Add one for c%stack, and one for the tag *)
                         then Specified (List.length fvcont + 2)
                         else framesize origf in
        let contargs = fvcont@["c%stack"] in
        let conttypes = List.map (fun v -> Context.find v context) contargs in
        ({i=expr.i; v=LetExp ("c%stack'",
          {i=unknown; v=ConstrExp (closname, List.map varval contargs, New)},
          {i=expr.i; v=AppExp (f, args@[varval "c%stack'"], appid)}, [])},
         [(fty,
           Matchrule (unknown, closname, contargs, false,
                      {i=unknown; v=AppExp (contname,
                                            List.map varval ("r"::contargs),
                                            make_appexpid ())}, New),
           TypCon (cont.i, closname, clossize, conttypes),
           ValDec (cont.i, contname, Unspecified,
                   make_arrowty (fty::conttypes) finaltype),
           FunctionDef (cont.i, contname, contvar::contargs, cont))])
    | LetExp (v,e1,e2,_) ->
        let vty = typeof (context, e1) in
        let context2 = Context.add v vty context in
        let (e2', f2) = exp (context2, e2, cont, contvar, tail, origf) in
        let (e1', f1) = exp (context, e1, e2', v, false, origf) in
        (e1', f1@f2)
    | SeqExp (e1,e2) ->
        let (e2', f2) = exp (context, e2, cont, contvar, tail, origf) in
        let (e1', f1) = exp (context, e1, e2', "c%unused", false, origf) in
        (e1', f1@f2)
    | IfExp (v,e1,e2) ->
        let (e1', f1) = exp (context, e1, cont, contvar, tail, origf) in
        let (e2', f2) = exp (context, e2, cont, contvar, tail, origf) in
        ({i=expr.i; v=IfExp (v,e1',e2')}, f1@f2)
    | MatchPExp (v,v1,v2,e) ->
        let t1,t2 = match (Context.find v context).v with
                      PairTyp(t1,t2) -> t1,t2
                    | _ -> failwith "pair match given non-pair variable"
         in
        let context2 = Context.add v2 t2 (Context.add v1 t1 context) in
        let (e',f) = exp (context2, e, cont, contvar, tail, origf) in
          ({i=expr.i; v=MatchPExp (v,v1,v2,e')}, f)
    | MatchExp (v,rules) ->
        let (rules', newfuns) =
          List.split (List.map (crule context cont contvar tail origf) rules) in
        ({i=expr.i; v=MatchExp (v,rules')}, List.flatten newfuns)
  and crule context cont contvar tail origf (Matchrule (i, c, bound, ro, e, d)) =
    let boundtys = (ConTab.find c contab).arg_types in
    let context' = List.fold_left (fun c (v,t) -> Context.add v t c)
      context (List.combine bound boundtys) in
    let (e', fns) = exp (context', e, cont, contvar, tail, origf) in
      (Matchrule (i, c, bound, ro, e', d), fns)
  in

  let cfun (FunctionDef (i,f,args,e)) =
    let (argtys, resultty) = fntys (fnty f) in
    let contextl = List.combine (args@["c%stack"]) (argtys@[stacktype]) in
    let context = List.fold_left (fun c (v,t) -> Context.add v t c)
                                 Context.empty contextl in
    let (e', newfns) = exp (context, e,
                            {i=unknown; v=
                                AppExp ("c%unwind_" ^ (string_typ resultty),
                                        [varval "c%result"; varval "c%stack"],
                                        make_appexpid ())},
                            "c%result", true, f) in
      (FunctionDef (i,f,args@["c%stack"],e'), newfns)
  in

  (* All of the original functions now take a stack and return the final type *)
  let trans_valdec orig =
    let (i, f, ty) =
      match orig with
      | ValDec (i,f,_,ty) -> (i,f,ty)
      | AnnValDec (i,f,_,_,ty,_)
      | FullValDec (i,f,_,_,ty,_,_) -> (i,f,strip_rich_typ ty)
    in
    let (argtys, _) = fntys ty in
      ValDec (i, f, Unspecified, make_arrowty (argtys@[stacktype]) finaltype)
  in

  (* Make the stack type *)
  let make_stack_type cons =
    let endc = TypCon (unknown, "C%end", Specified 0, []) in
      TypDec (unknown, [], "c%stack", cons@[endc])
  in

  (* Make the unwind functions.  Note that we have to be careful to ignore
     the source position information when comparing types. *)
  let make_unwind (ty, rules) =
    let ty' = {i=unknown; v=ty} in
    let rules' =
      if ty = finaltype.v
      then (Matchrule (unknown, "C%end", [], false,
                       {i=unknown; v=ValueExp (varval "r")}, New))::rules
      else rules
    in
      (ValDec (unknown, "c%unwind_" ^ (string_typ ty'), Unspecified,
               make_arrowty [ty'; stacktype] finaltype),
       FunctionDef (unknown, "c%unwind_" ^ (string_typ ty'), ["r"; "c%stack1"],
                    {i=unknown; v=MatchExp("c%stack1", rules')}))
  in

  let partition_rules conts =
    let t = Hashtbl.create 31 in
    let enter (ty,r,_,_,_) =
      let old = try Hashtbl.find t ty.v with Not_found -> [] in
        Hashtbl.replace t ty.v (r::old)
    in
      Hashtbl.add t finaltype.v [];
      List.iter enter conts;
      Hashtbl.fold (fun ty rs t -> (ty,rs)::t) t []
  in

  let make_unwinds conts =
    let rules_by_type = partition_rules conts in
    let unwinds = List.map make_unwind rules_by_type in
      List.split unwinds
  in

  let new_maindef =
    let mki e = {i=unknown; v=e} in
    let FunctionDef (i,f,args,e) =
        (List.find (fun (FunctionDef (_,f,_,_)) -> f = mainfun)
          (List.flatten fundefs)) in
      FunctionDef (unknown, "c%main", args,
                   mki (LetExp ("s", mki (ConstrExp ("C%end", [], New)),
                     mki (AppExp (f, List.map varval (args@["s"]),
                            make_appexpid ())), [])))
  in

  let new_maindec =
    let ValDec (i, f, _, ty) = get_val_for mainfun valdecs in
      ValDec (unknown, "c%main", Unspecified, ty)
  in

  (* Put it all together *)
  let (trans_fundefs, conts) =
    List.fold_left (fun (tfns,nfns) f ->
                      let (tfn,nfn) = cfun f in
                        (tfn::tfns,nfn@nfns)) ([],[]) (List.flatten fundefs) in
  let (cont_cons, cont_valdecs, cont_fundefs) =
    List.fold_left (fun (cs,vs,fs) (_,_,c,v,f) -> (c::cs, v::vs, f::fs))
      ([],[],[]) conts in
  let trans_valdecs = List.map trans_valdec valdecs in
  let stacktypedec = make_stack_type cont_cons in
  let tydecs' = if (!Argument.the_options).Argument.cpsnodata
                  then List.map zero_tydec tydecs
                  else tydecs
  in
  let (unwind_valdecs, unwind_fundefs) = make_unwinds conts in
    Program (errinfo, stacktypedec::tydecs',
             new_maindec :: unwind_valdecs @ cont_valdecs @ trans_valdecs,
             [new_maindef :: unwind_fundefs @ cont_fundefs @ trans_fundefs])
