open Support
open Support.Error
open Syntax

module VarSet = Set.Make(struct type t = string let compare = compare end)
module StringMap = Map.Make(struct type t = string let compare = compare end)

let just s = (s,VarSet.cardinal s)
let maxfv (s, m) = (s,max m (VarSet.cardinal s))

let fv_val_set v = List.fold_left (fun a b -> VarSet.add b a) VarSet.empty (Syntax.fv_val v)

let rec fv_max expr =
  match expr.v with
  |  ValueExp v
      -> just (fv_val_set v)
  | ConstrExp (_,args,_)
  | AppExp (_,args,_)
      -> just (List.fold_left (fun s a -> VarSet.union (fv_val_set a) s) VarSet.empty args)
  | LetExp (bound, e1, e2, _)
      -> let (fv1,m1), (fv2,m2) = fv_max e1, fv_max e2 in
         let filteredfv2 = VarSet.remove bound fv2 in
         (* need to remember fv for e2 while executing e1 *)
         maxfv (VarSet.union fv1 filteredfv2,
                max (VarSet.cardinal filteredfv2 + m1) m2)
  | SeqExp (e1,e2)
      -> let (fv1,m1), (fv2,m2) = fv_max e1, fv_max e2 in
         maxfv (VarSet.union fv1 fv2, max (VarSet.cardinal fv2 + m1) m2)
  | IfExp (v,e1,e2)
      -> let fv0, (fv1,m1), (fv2,m2) = fv_val_set v, fv_max e1, fv_max e2 in
         maxfv (VarSet.union fv0 (VarSet.union fv1 fv2), max m1 m2)
  | MatchExp (v, rules)
      -> maxfv (List.fold_left fv_rule_max (VarSet.add v VarSet.empty, 0) rules)
and fv_rule_max (fvs,m1) (Matchrule (_,_,bound,_,e,_)) =
  let fv,m = fv_max e in
  let fv' = VarSet.filter (fun v -> not (List.mem v bound)) fv in
  (VarSet.union fvs fv', max m m1)

let framesize =
  function FunctionDef(_,_,args,expr) ->
    let _,fvmax = fv_max expr in
    max fvmax (List.length args)

let estimate_funs funblocks =
  List.fold_left
    (List.fold_left (fun m ->
                       function (FunctionDef(_,f,_,_) as fd) ->
                         StringMap.add f (framesize fd) m))
    StringMap.empty funblocks

let update framesizes =
  let upfn v =
    try
      match v with
        ValDec (i,f,_,t) -> ValDec (i,f,Specified (StringMap.find f framesizes),t)
      | AnnValDec (i,f,_,n,t,n') ->
          AnnValDec (i,f,Specified (StringMap.find f framesizes),n,t,n')
      | FullValDec (i,f,_,n,t,n',c) -> 
          FullValDec (i,f,Specified (StringMap.find f framesizes),n,t,n',c)
    with Not_found -> v
  in List.map upfn

let estimate_program =
  function Program(i,tys,vals,funblocks) ->
    let framesizes = estimate_funs funblocks in
    Program(i,tys,update framesizes vals,funblocks)
